/*-----------------------------------------------------------------------------
 * $RCSfile: Transport.c,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if (!defined(_WIN32)) || defined(_DEBUG)
static  const   char    CVS_ID[]        =       "$Id: Transport.c,v 1.21 2001/12/18 02:00:55 bandou Exp $";
#endif  /* _DEBUG */

#include "Transport.h"

#include <OpenSOAP/ClientSocket.h>

#include <string.h>
#include <ctype.h>

/*
 */

#if defined(NONGCC_WIN32)
# define STRING_WFORMAT L"%hs"
#else /* !NONGCC_WIN32 */
# define STRING_WFORMAT L"%s"
#endif /* !NONGCC_WIN32 */

static
const char NULL_CHAR = '\0';

static
const unsigned char
HTTP_LINE_TERM[] = "\r\n";

#define HTTP_LINE_TERM_SZ (sizeof(HTTP_LINE_TERM) - 1)

static
const unsigned char
HTTP_HEADER_TERM[] = "\r\n\r\n";

#define HTTP_HEADER_TERM_SZ (sizeof(HTTP_HEADER_TERM) - 1)

static
const unsigned char *
HTTP_LINE_TERM_END = HTTP_LINE_TERM + HTTP_LINE_TERM_SZ;

/*
=begin
= OpenSOAP ConnectInfo class
=end
 */

struct tagOpenSOAPConnectInfo {
    char    *hostname;
    char    *protocol;
    char    *port;
    char    *location;
};

static
const char
URI_PRE_HOSTNAME[] = "://";

static
const char
URI_DELIMITER[] = "/";

static
const char
URI_PORT_DELIMITER[] = ":";

static
int
OpenSOAPConnectInfoReleaseMembers(OpenSOAPConnectInfoPtr
                                  /* [in, out] */ c_info) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;
    
    if (c_info) {
        free(c_info->hostname); c_info->hostname = NULL;
        free(c_info->protocol); c_info->protocol = NULL;
        free(c_info->port);     c_info->port = NULL;
        free(c_info->location); c_info->location = NULL;

        ret = OPENSOAP_NO_ERROR;
    }
    
    return ret;
}

static
int
OpenSOAPConnectInfoFree(OpenSOAPConnectInfoPtr
                        /* [in, out] */ c_info) {
    int ret = OpenSOAPConnectInfoReleaseMembers(c_info);

    if (OPENSOAP_SUCCEEDED(ret)) {
        free(c_info);
    }

    return ret;
}

static
int
OpenSOAPConnectInfoInitialize(OpenSOAPConnectInfoPtr
                              /* [in, out] */ c_info) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;
    
    if (c_info) {
        c_info->hostname = NULL;
        c_info->protocol = NULL;
        c_info->port = NULL;
        c_info->location = NULL;

        ret = OPENSOAP_NO_ERROR;
    }
    
    return ret;
}

static
int
OpenSOAPConnectInfoCreate(OpenSOAPConnectInfoPtr * /* [out] */ c_info) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;
    
    if (c_info) {
        *c_info = malloc(sizeof(OpenSOAPConnectInfo));
        if (*c_info) {
            ret = OpenSOAPConnectInfoInitialize(*c_info);
            if (OPENSOAP_FAILED(ret)) {
                free(*c_info);
                *c_info = NULL;
            }
        }
    }
    
    return ret;
}

#if 1
# define OpenSOAPConnectInfoRelease(c_info) OpenSOAPConnectInfoFree(c_info)
#else
static
int
OpenSOAPConnectInfoRelease(OpenSOAPConnectInfoPtr /* [in, out] */ c_info) {
    int ret = OpenSOAPConnectInfoFree(c_info);

    return ret;
}
#endif

static
int
OpenSOAPConnectInfoSetURL(OpenSOAPConnectInfoPtr /* [in, out] */ c_info,
                          const char * /* [in] */ url) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;
    
    if (c_info && url && *url) {
        const char *protocol_begin = url;
        const char *protocol_end = NULL;
        const char *hostname_begin = url;
        const char *hostname_end = NULL;
        const char *port_begin = NULL;
        const char *port_end = NULL;
        const char *location_begin = NULL;
        size_t hostname_len = 0;
        size_t protocol_len = 0;
        size_t port_len = 0;
        size_t location_len = 0;

        protocol_end = strstr(protocol_begin, URI_PRE_HOSTNAME);
        if (protocol_end) {
            hostname_begin = protocol_end + strlen(URI_PRE_HOSTNAME);

            hostname_end = strstr(hostname_begin, URI_DELIMITER);
            location_begin = hostname_end;

            if (!hostname_end) {
                hostname_end = hostname_begin + strlen(hostname_begin);
                location_begin = URI_DELIMITER;
            }
        
            port_begin = strstr(hostname_begin, URI_PORT_DELIMITER);
            if (port_begin) {
                if (port_begin < hostname_end) {
                    port_end = hostname_end;
                    hostname_end = port_begin;
                    ++port_begin;
                }
                else {
                    port_begin = NULL;
                }
            }

            if (!port_begin) {
                port_begin = protocol_begin;
                port_end   = protocol_end;
            }

            hostname_len = hostname_end - hostname_begin;
            protocol_len = protocol_end - protocol_begin;
            port_len = port_end - port_begin;
            location_len = strlen(location_begin);
        
            OpenSOAPConnectInfoReleaseMembers(c_info);
        
            c_info->hostname = malloc(hostname_len + 1);
            c_info->protocol = malloc(protocol_len + 1);
            c_info->port = malloc(port_len + 1);
            c_info->location = malloc(location_len + 1);

            if (c_info->hostname && c_info->protocol
                && c_info->port && c_info->location) {
                strncpy(c_info->hostname,
                        hostname_begin,
                        hostname_len);
                c_info->hostname[hostname_len] = '\0';
                strncpy(c_info->protocol,
                        protocol_begin,
                        protocol_len);
                c_info->protocol[protocol_len] = '\0';
                strncpy(c_info->port,
                        port_begin,
                        port_len);
                c_info->port[port_len] = '\0';
                strncpy(c_info->location,
                        location_begin,
                        location_len);
                c_info->location[location_len] = '\0';

                ret = OPENSOAP_NO_ERROR;
            }
            else {
                OpenSOAPConnectInfoReleaseMembers(c_info);
                ret = OPENSOAP_MEM_BADALLOC;
            }
        }
    }
    
    return ret;
}

static
int
OpenSOAPConnectInfoGetHTTPPOSTRequest(OpenSOAPConnectInfoPtr
                                      /* [in, out] */ c_info,
                                      const char * /* [in] */ http_ver,
                                      OpenSOAPByteArrayPtr
                                      /* [out] */ http_hdr) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;
    static const unsigned char DEFAULT_HTTP_VER[] = "1.0";
    static const wchar_t POST_REQUEST_FORMAT[]
        =
        L"POST "
        STRING_WFORMAT  /* Request */
        L" HTTP/"
        STRING_WFORMAT  /* HTTP version */
        STRING_WFORMAT; /* line term */

    if (!http_ver) {
        http_ver = DEFAULT_HTTP_VER;
    }
    if (c_info && c_info->location && *(c_info->location) && http_hdr) {
        OpenSOAPStringPtr tmp = NULL;
        ret = OpenSOAPStringCreate(&tmp);

        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPStringFormatWC(tmp,
                                         POST_REQUEST_FORMAT,
                                         c_info->location,
                                         http_ver,
                                         HTTP_LINE_TERM);
            if (OPENSOAP_SUCCEEDED(ret)) {
                ret = OpenSOAPStringGetCharEncodingString(tmp,
                                                          "US-ASCII",
                                                          http_hdr);
            }

            OpenSOAPStringRelease(tmp);
        }
    }

    return ret;
}

static
int
OpenSOAPConnectInfoIsSameProtocol(OpenSOAPConnectInfoPtr
                                  /* [in] */ c_info,
                                  const char * /* [in] */ protocol,
                                  int * /* [out] */ is_same) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (c_info && protocol && *protocol && is_same) {
        *is_same = (strcmp(protocol, c_info->protocol) == 0);
        ret = OPENSOAP_NO_ERROR;
    }

    return ret;
}

static
int
OpenSOAPConnectInfoCreateClientSocketStream(OpenSOAPConnectInfoPtr
                                            /* [in] */ c_info,
                                            OpenSOAPStreamPtr *
                                            /* [out] */ s) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (c_info && s) {
        ret = OpenSOAPClientSocketCreate((OpenSOAPClientSocketPtr *)s);

        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPClientSocketOpen((OpenSOAPClientSocketPtr)*s,
                                           c_info->hostname,
                                           c_info->port);
            if (OPENSOAP_FAILED(ret)) {
                OpenSOAPStreamRelease(*s);
                *s = NULL;
            }
        }
    }

    return ret;
}

#include <OpenSOAP/StringHash.h>

/*
=begin
= OpenSOAP TransportHeader class
=end
 */

struct tagOpenSOAPTransportHeader {
    OpenSOAPObject  super;
    OpenSOAPStringHashPtr   headerMap;
    OpenSOAPByteArrayPtr    charset;
    size_t                  contentLength;
};

static
const unsigned char
CONTENT_LENGTH_NAME[] = "Content-Length";

static
const unsigned char
CONTENT_TYPE_NAME[] = "Content-Type";

static
const unsigned char
SOAPACTION_NAME[] = "SOAPAction";

static
const size_t
SOAPACTION_NAME_SZ = sizeof(SOAPACTION_NAME) - 1;

/*
=begin
--- function#OpenSOAPTransportHeaderMapValueFree(value, noused)
    TransportHeader  header_map ̒lJB
    
    :Parameters
      :void * [in] ((|value|))
        TransportHeader  header_map ̒l
      :void * [in, out] ((|noused|))
        no use option.
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
static
int
OpenSOAPTransportHeaderMapValueFree(void * /* [in] */ value,
                                 void * /* [in, out] */ noused) {
    int ret = OPENSOAP_NO_ERROR;

    if (value) {
        free(value);
    }

    return ret;
}

static
int
OpenSOAPTransportHeaderClear(OpenSOAPTransportHeaderPtr /* [in, out] */ th) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (th) {
        ret = OpenSOAPStringHashApplyToValues(th->headerMap,
                                              OpenSOAPTransportHeaderMapValueFree,
                                              NULL);
        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPStringHashClear(th->headerMap);
            if (OPENSOAP_SUCCEEDED(ret)) {
                th->contentLength = 0;
                ret = OpenSOAPByteArraySetData(th->charset,
                                               &NULL_CHAR,
                                               1);
            }
        }
    }

    return ret;
}

static
int
OpenSOAPTransportHeaderReleaseMembers(OpenSOAPTransportHeaderPtr
                                      /* [in, out] */ th) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (th) {
        ret = OpenSOAPStringHashApplyToValues(th->headerMap,
                                              OpenSOAPTransportHeaderMapValueFree,
                                              NULL);
        if (OPENSOAP_SUCCEEDED(ret)) {
            OpenSOAPStringHashRelease(th->headerMap);
            OpenSOAPByteArrayRelease(th->charset);
            ret = OpenSOAPObjectReleaseMembers((OpenSOAPObjectPtr)th);
        }
    }

    return ret;
}

static
int
OpenSOAPTransportHeaderFree(OpenSOAPObjectPtr /* [in, out] */ obj) {
    int ret = OpenSOAPTransportHeaderReleaseMembers((OpenSOAPTransportHeaderPtr)obj);
    if (OPENSOAP_SUCCEEDED(ret)) {
        free(obj);
    }

    return ret;
}

static
int
OpenSOAPTransportHeaderInitialize(OpenSOAPTransportHeaderPtr
                                  /* [in, out] */ th) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (th) {
        OpenSOAPStringHashPtr hdr_map = NULL;
        th->headerMap = NULL;
        th->charset = NULL;
        th->contentLength = 0;
        ret = OpenSOAPStringHashCreate(&hdr_map);
        if (OPENSOAP_SUCCEEDED(ret)) {
            OpenSOAPByteArrayPtr chset = NULL;
            ret = OpenSOAPByteArrayCreateWithData(&NULL_CHAR, 1, &chset);
            if (OPENSOAP_SUCCEEDED(ret)) {
                ret = OpenSOAPObjectInitialize((OpenSOAPObjectPtr)th,
                                               OpenSOAPTransportHeaderFree);
                if (OPENSOAP_SUCCEEDED(ret)) {
                    th->headerMap = hdr_map;
                    th->charset = chset;
                    th->contentLength = 0;
                }
                if (OPENSOAP_FAILED(ret)) {
                    OpenSOAPByteArrayRelease(chset);
                }
            }
            if (OPENSOAP_FAILED(ret)) {
                OpenSOAPStringHashRelease(hdr_map);
            }
        }
    }

    return ret;
}

static
int
OpenSOAPTransportHeaderCreate(OpenSOAPTransportHeaderPtr *
                              /* [out] */ th) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (th) {
        ret = OPENSOAP_MEM_BADALLOC;
        *th = malloc(sizeof(OpenSOAPTransport));
        if (*th) {
            ret = OpenSOAPTransportHeaderInitialize(*th);
            if (OPENSOAP_FAILED(ret)) {
                free(*th);
                *th = NULL;
            }
        }
    }

    return ret;
}

static
int
OpenSOAPTransportHeaderRelease(OpenSOAPTransportHeaderPtr 
                               /* [in, out] */ th) {
    int ret = OpenSOAPObjectRelease((OpenSOAPObjectPtr)th);

    return ret;
}

static
int
OpenSOAPTransportHeaderSetHeader(OpenSOAPTransportHeaderPtr
                                 /* [in, out] */ th,
                                 const char * /* [in] */ name,
                                 size_t /* [in] */ name_sz,
                                 const char * /* [in] */ val,
                                 size_t /* [in] */ val_sz) {
    static const char CHARSET_EQUAL[] = "charset=";
    static const size_t CHARSET_EQUAL_SIZE = sizeof(CHARSET_EQUAL) - 1;
    static const char QUOTATIONS[] = "\'\"";

    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (th && name && name_sz) {
        char *name_mbs = malloc(name_sz + 1);
        ret = OPENSOAP_MEM_BADALLOC;
        if (name_mbs) {
            char *val_mbs = malloc(val_sz + 1);
            if (val_mbs) {
                OpenSOAPByteArrayPtr    charset = NULL;
                int is_content_length = 0;
                size_t content_length = 0;
                ret = OPENSOAP_NO_ERROR;
                memcpy(name_mbs, name, name_sz);
                name_mbs[name_sz] = '\0';
                memcpy(val_mbs, val, val_sz);
                val_mbs[val_sz]  = '\0';
                if (strcmp(CONTENT_TYPE_NAME, name_mbs) == 0) {
                    char *val_mbs_end = val_mbs + val_sz;
                    char *charset_beg = strstr(val_mbs,
                                               CHARSET_EQUAL);
                    if (charset_beg) {
                        ret = OpenSOAPByteArrayCreate(&charset);
                        if (OPENSOAP_SUCCEEDED(ret)) {
                            char tmp = NULL_CHAR;
                            size_t charset_sz = 0;
                            char *c_i = NULL;
                            const char *q_i = QUOTATIONS;

                            for (charset_beg += CHARSET_EQUAL_SIZE;
                                 *q_i && *q_i != *charset_beg; ++q_i) {
                            }
                            if (*q_i) {
                                ++charset_beg;
                                for (c_i = charset_beg;
                                     c_i != val_mbs_end && *c_i != *q_i;
                                     ++c_i) {
                                }
                            }
                            else {
                                for (c_i = charset_beg;
                                     c_i != val_mbs_end && !isspace(*c_i);
                                     ++c_i) {
                                }
                            }
                            tmp = *c_i;
                            *c_i = NULL_CHAR;
                            charset_sz = c_i - charset_beg + 1;
                            ret = OpenSOAPByteArraySetData(charset,
                                                           charset_beg,
                                                           charset_sz);
                            *c_i = tmp;
                            if (OPENSOAP_FAILED(ret)) {
                                OpenSOAPByteArrayRelease(charset);
                                charset = NULL;
                            }
                        }
                    }
                }
                else if (strcmp(CONTENT_LENGTH_NAME, name_mbs)) {
                    content_length
                        = (name_mbs) ? atol(name_mbs) : 0;
                    is_content_length = 1;
                }
                if (OPENSOAP_SUCCEEDED(ret)) {
                    void *old_val = NULL;
                    ret = OpenSOAPStringHashGetValueMB(th->headerMap,
                                                       name_mbs,
                                                       &old_val);
                    if (OPENSOAP_SUCCEEDED(ret)) {
                        free(old_val);
                    }
                    ret = OpenSOAPStringHashSetValueMB(th->headerMap,
                                                       name_mbs,
                                                       val_mbs);
                }
                if (OPENSOAP_SUCCEEDED(ret)) {
                    if (charset) {
                        OpenSOAPByteArrayRelease(th->charset);
                        th->charset = charset;
                    }
                    if (is_content_length) {
                        th->contentLength = content_length;
                    }
                }
                else {
                    free(val_mbs);
                }
            }
            free(name_mbs);
        }
    }

    return ret;
}

static
int
OpenSOAPTransportHeaderSetContentLength(OpenSOAPTransportHeaderPtr 
                                         /* [in, out] */ th,
                                        size_t /* [in] */ content_length) {
    
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (th) {
        OpenSOAPByteArrayPtr tmp_buf = NULL;
        ret = OpenSOAPByteArrayCreate(&tmp_buf);
        if (OPENSOAP_SUCCEEDED(ret)) {
            OpenSOAPStringPtr tmp_str = NULL;
            ret = OpenSOAPStringCreate(&tmp_str);
            if (OPENSOAP_SUCCEEDED(ret)) {
                ret = OpenSOAPStringFormatWC(tmp_str,
                                             L"%ld",
                                             content_length);
                if (OPENSOAP_SUCCEEDED(ret)) {
                    ret = OpenSOAPStringGetCharEncodingString(tmp_str,
                                                              NULL,
                                                              tmp_buf);
                    if (OPENSOAP_SUCCEEDED(ret)) {
                        const unsigned char *c_l_beg = NULL;
                        size_t c_l_sz = 0;
                        ret = OpenSOAPByteArrayGetBeginSizeConst(tmp_buf,
                                                                 &c_l_beg,
                                                                 &c_l_sz);
                        
                        if (OPENSOAP_SUCCEEDED(ret)) {
                            ret = OpenSOAPTransportHeaderSetHeader(th,
                                                                   CONTENT_LENGTH_NAME,
                                                                   sizeof(CONTENT_LENGTH_NAME) - 1,
                                                                   c_l_beg,
                                                                   c_l_sz);
                        
                        }
                    }
                }
                OpenSOAPStringRelease(tmp_str);
            }
            OpenSOAPByteArrayRelease(tmp_buf);
        }
    }

    return ret;
}

#if 0
static
int
OpenSOAPTransportHeaderSetCharset(OpenSOAPTransportHeaderPtr 
                                  /* [in, out] */ th,
                                  const unsigned char *
                                  /* [in] */ charset) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (th && charset && *charset) {
        size_t charset_sz = strlen(charset);
        const unsigned char *content_type = NULL;
        ret = OpenSOAPStringHashGetValueMB(th->headerMap,
                                           CONTENT_TYPE_NAME,
                                           &content_type);
        if (OPENSOAP_SUCCEEDED(ret)) {
            
        }
        else { /* no content type */
            
        }
    }

    return ret;
}
#endif

static
int
OpenSOAPTransportHeaderSetFromHTTPHeader(OpenSOAPTransportHeaderPtr 
                                         /* [in, out] */ th,
                                         OpenSOAPByteArrayPtr
                                         /* [in] */ http_hdr,
                                         int * /* [out] */ http_code) {
    static const int HTTP_OK_CODE = 200;
    static const unsigned char HEADER_DELM = ':';
    static const unsigned char HTTP_RESPONSE_PREFIX[] = "HTTP/";
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (th && http_hdr) {
        int dummy_http_code = 0;
        const unsigned char *http_hdr_beg = NULL;
        const unsigned char *http_hdr_end = NULL;
        if (!http_code) {
            http_code = &dummy_http_code;
        }
        *http_code = HTTP_OK_CODE;
        ret = OpenSOAPByteArrayGetBeginEndConst(http_hdr,
                                                &http_hdr_beg,
                                                &http_hdr_end);
        if (OPENSOAP_SUCCEEDED(ret)) {
            const unsigned char *d = NULL;
            const unsigned char *l_t_i = HTTP_LINE_TERM;
            const unsigned char *i = http_hdr_beg;
            const unsigned char *l_beg = i;
            int is_checked_res = 0;
            for (; i != http_hdr_end; ++i) {
                if (*i == *l_t_i) {
                    ++l_t_i;
                }
                else {
                    l_t_i = HTTP_LINE_TERM;
                }
                if (*i == HEADER_DELM) {
                    d = i;
                }
                if (l_t_i == HTTP_LINE_TERM_END) {
                    const unsigned char *l_end
                        = i - (HTTP_LINE_TERM_END - HTTP_LINE_TERM) + 1;
                    if (l_beg == l_end) {
                        break;
                    }
                    if (!is_checked_res) {
                        if (!d) {
                            if (strncmp(l_beg,
                                        HTTP_RESPONSE_PREFIX,
                                        sizeof(HTTP_RESPONSE_PREFIX) - 1)
                                == 0) {
                                const unsigned char *h_r_i
                                    = l_beg
                                    + sizeof(HTTP_RESPONSE_PREFIX) - 1;
                                for (; h_r_i != l_end
                                         && (isdigit(*h_r_i)
                                             || *h_r_i == '.'); ++h_r_i) {
                                }
                                for (; h_r_i != l_end && isspace(*h_r_i);
                                     ++h_r_i) {
                                }
                                if (h_r_i != l_end) {
                                    for (*http_code = 0;
                                         h_r_i != l_end && isdigit(*h_r_i);
                                         ++h_r_i) {
                                        *http_code *= 10;
                                        *http_code += (*h_r_i - '0');
                                    }
                                }
                            }
                        }
                        is_checked_res = 1;
                    }
                    if (d) {
                        const unsigned char *v = d;

                        for (; v != l_end && isspace(*v); ++v) {
                        }
                        ret = OpenSOAPTransportHeaderSetHeader(th,
                                                               l_beg,
                                                               d - l_beg,
                                                               v,
                                                               l_end - v);
                        if (OPENSOAP_FAILED(ret)) {
                            break;
                        }
                    }
                    
                    l_t_i = HTTP_LINE_TERM;
                    l_beg = i + 1;
                    d = NULL;
                }
            }
        }
    }

    return ret;
}

static
int
OpenSOAPByteArrayAppendHTTPHeader(OpenSOAPByteArrayPtr
                                  /* [in, out] */ http_hdr,
                                  OpenSOAPStringPtr /* [in] */ name,
                                  const char * /* [in] */ val) {
    static const unsigned char HTTP_HEADER_DELM[] = ": ";
    static const size_t HTTP_HEADER_DELM_SZ
        = sizeof(HTTP_HEADER_DELM) - 1;
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (http_hdr && name) {
        OpenSOAPByteArrayPtr line_buf = NULL;
        ret = OpenSOAPByteArrayCreate(&line_buf);

        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPStringGetCharEncodingString(name,
                                                      "US-ASCII",
                                                      line_buf);
            if (OPENSOAP_SUCCEEDED(ret)) {
                ret = OpenSOAPByteArrayAppend(line_buf,
                                              HTTP_HEADER_DELM,
                                              HTTP_HEADER_DELM_SZ);
                if (OPENSOAP_SUCCEEDED(ret)) {
                    size_t val_sz = val ? strlen(val) : 0;

                    if (val_sz) {
                        ret = OpenSOAPByteArrayAppend(line_buf,
                                                      val,
                                                      val_sz);
                    }
                    if (OPENSOAP_SUCCEEDED(ret)) {
                        ret = OpenSOAPByteArrayAppend(line_buf,
                                                      HTTP_LINE_TERM,
                                                      HTTP_LINE_TERM_SZ);
                    }
                }
            }

            if (OPENSOAP_SUCCEEDED(ret)) {
                const unsigned char *line_beg = NULL;
                size_t line_sz = 0;
                ret = OpenSOAPByteArrayGetBeginSizeConst(line_buf,
                                                         &line_beg,
                                                         &line_sz);
                if (OPENSOAP_SUCCEEDED(ret)) {
                    ret = OpenSOAPByteArrayAppend(http_hdr,
                                                  line_beg,
                                                  line_sz);
                }
            }
            OpenSOAPByteArrayRelease(line_buf);
        }
    }

    return ret;
}

static
int
OpenSOAPTransportHeaderAppendToHTTPHeader(OpenSOAPTransportHeaderPtr
                                          /* [in] */ th,
                                          OpenSOAPByteArrayPtr
                                          /* [out] */ http_hdr) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (th && http_hdr) {
        void *soapaction_val = NULL;
        ret = OpenSOAPStringHashGetValueMB(th->headerMap,
                                           SOAPACTION_NAME,
                                           &soapaction_val);
        if (OPENSOAP_FAILED(ret)) {
            ret = OpenSOAPStringHashSetValueMB(th->headerMap,
                                               SOAPACTION_NAME,
                                               soapaction_val);
        }
        if (OPENSOAP_SUCCEEDED(ret)) {
            size_t map_sz = 0;
            ret = OpenSOAPStringHashGetSize(th->headerMap, &map_sz);
            if (OPENSOAP_SUCCEEDED(ret)) {
                OpenSOAPStringPtr *map_keys
                    = malloc(sizeof(OpenSOAPStringPtr) * map_sz);
                if (!map_keys) {
                    ret = OPENSOAP_MEM_BADALLOC;
                }
                else {
                    ret = OpenSOAPStringHashGetKeys(th->headerMap,
                                                    &map_sz,
                                                    map_keys);
                    if (OPENSOAP_SUCCEEDED(ret)) {
                        OpenSOAPStringPtr *map_keys_end
                            = map_keys + map_sz;
                        OpenSOAPStringPtr *map_keys_i
                            = map_keys;
                    
                        for (map_keys_i = map_keys;
                             OPENSOAP_SUCCEEDED(ret)
                                 && map_keys_i != map_keys_end;
                             ++map_keys_i) {
                            const char *val = NULL;
                            ret = OpenSOAPStringHashGetValue(th->headerMap,
                                                             *map_keys_i,
                                                             (void **)&val);
                            if (OPENSOAP_SUCCEEDED(ret)) {
                                ret = OpenSOAPByteArrayAppendHTTPHeader(http_hdr,
                                                                        *map_keys_i,
                                                                        val);
                            }
                        }
                        
                        for (map_keys_i = map_keys;
                             map_keys_i != map_keys_end;
                             ++map_keys_i) {
                            OpenSOAPStringRelease(*map_keys_i);
                        }

                        if (OPENSOAP_SUCCEEDED(ret)) {
                            ret = OpenSOAPByteArrayAppend(http_hdr,
                                                          HTTP_LINE_TERM,
                                                          HTTP_LINE_TERM_SZ);
                        }
                    }

                    free(map_keys);
                }
            }
        }
    }

    return ret;
}

static
int
OpenSOAPTransportHeaderGetCharsetBeginConst(OpenSOAPTransportHeaderPtr
                                            /* [in] */ th,
                                            const unsigned char **
                                            /* [out] */ charset_beg) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (th && charset_beg) {
        ret = OpenSOAPByteArrayBeginConst(th->charset,
                                          charset_beg);
    }

    return ret;
}

static
int
OpenSOAPTransportHeaderIsContentTypeXML(/* [in] */ OpenSOAPTransportHeaderPtr th,
										/* [out] */ int *is_xml) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (th && is_xml) {
		const char *content_type = NULL;
		*is_xml = 0;
		ret = OpenSOAPStringHashGetValueMB(th->headerMap,
										   CONTENT_TYPE_NAME,
										   (void **)&content_type);
		if (OPENSOAP_SUCCEEDED(ret)) {
			static const char CONTENT_TYPE_TEXT_XML[] = "text/xml";
			const char *text_xml_beg = strstr(content_type,
											  CONTENT_TYPE_TEXT_XML);
			if (text_xml_beg) {
				*is_xml = 1;
			}
		}
		ret = OPENSOAP_NO_ERROR;
    }

    return ret;
}

static
int
OpenSOAPTransportHeaderGetContentLength(OpenSOAPTransportHeaderPtr
                                        /* [in] */ th,
                                        size_t *
                                        /* [out] */ content_length) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (th && content_length) {
        *content_length = th->contentLength;
        ret = OPENSOAP_NO_ERROR;
    }
    
    return ret;
}


/*
=begin
= OpenSOAP Transport class
=end
 */

/*
=begin
--- function#OpenSOAPTransportReleaseMembers(t)
    gX|[g̃o[JB
    
    :Parameters
      :OpenSOAPTransportPtr * [out] ((|t|))
        OpenSOAP Transport pointer
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
static
int
OpenSOAPTransportReleaseMembers(OpenSOAPTransportPtr t) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t) {
        OpenSOAPTransportHeaderRelease(t->requestHeaders);
        OpenSOAPTransportHeaderRelease(t->responseHeaders);
        OpenSOAPStreamRelease(t->transportStream);
        OpenSOAPByteArrayRelease(t->serviceEndPoint);
        OpenSOAPConnectInfoRelease(t->serviceConnectInfo);
        ret = OpenSOAPObjectReleaseMembers((OpenSOAPObjectPtr)t);
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPTransportFree(t)
    gX|[gJB
    
    :Parameters
      :OpenSOAPTransportPtr * [out] ((|t|))
        OpenSOAP Transport pointer
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
static
int
OpenSOAPTransportFree(OpenSOAPObjectPtr obj) {
    OpenSOAPTransportPtr t = (OpenSOAPTransportPtr)obj;
    int ret = OpenSOAPTransportReleaseMembers(t);

    if (OPENSOAP_SUCCEEDED(ret)) {
        free(obj);
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPTransportInitialize(t)
    gX|[g̃o[B
    
    :Parameters
      :OpenSOAPTransportPtr * [out] ((|t|))
        OpenSOAP Transport pointer
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
static
int
OpenSOAPTransportInitialize(OpenSOAPTransportPtr t) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t) {
        t->requestHeaders  = NULL;
        t->responseHeaders = NULL;
        t->transportStream = NULL;
        t->serviceEndPoint = NULL;
        t->serviceConnectInfo = NULL;
        ret = OpenSOAPObjectInitialize((OpenSOAPObjectPtr)t,
                                       OpenSOAPTransportFree);
        if (OPENSOAP_SUCCEEDED(ret)) {
            OpenSOAPTransportHeaderPtr req_hdr = NULL;
            ret = OpenSOAPTransportHeaderCreate(&req_hdr);
            if (OPENSOAP_SUCCEEDED(ret)) {
                OpenSOAPTransportHeaderPtr res_hdr = NULL;
                ret = OpenSOAPTransportHeaderCreate(&res_hdr);
                if (OPENSOAP_SUCCEEDED(ret)) {
                    OpenSOAPByteArrayPtr s_e_p = NULL;
                    ret = OpenSOAPByteArrayCreate(&s_e_p);
                    if (OPENSOAP_SUCCEEDED(ret)) {
                        OpenSOAPConnectInfoPtr c_info = NULL;
                        ret = OpenSOAPConnectInfoCreate(&c_info);
                        if (OPENSOAP_SUCCEEDED(ret)) {
                            t->requestHeaders  = req_hdr;
                            t->responseHeaders = res_hdr;
                            t->serviceEndPoint = s_e_p;
                            t->serviceConnectInfo = c_info;
                        }
                        if (OPENSOAP_FAILED(ret)) {
                            OpenSOAPByteArrayRelease(s_e_p);
                        }
                    }
                    if (OPENSOAP_FAILED(ret)) {
                        OpenSOAPTransportHeaderRelease(res_hdr);
                    }
                }
                if (OPENSOAP_FAILED(ret)) {
                    OpenSOAPTransportHeaderRelease(req_hdr);
                }
            }
            if (OPENSOAP_FAILED(ret)) {
                OpenSOAPObjectReleaseMembers((OpenSOAPObjectPtr)t);
            }
        }
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPTransportCreate(t)
    gX|[g𐶐B
    
    :Parameters
      :OpenSOAPTransportPtr * [out] ((|t|))
        OpenSOAP Transport pointer
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportCreate(OpenSOAPTransportPtr * /* [out] */ t) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    ret = OPENSOAP_PARAMETER_BADVALUE;
    if (t) {
        ret = OPENSOAP_MEM_BADALLOC;
        *t = malloc(sizeof(OpenSOAPTransport));
        if (*t) {
            ret = OpenSOAPTransportInitialize(*t);
            if (OPENSOAP_FAILED(ret)) {
                free(*t);
            }
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportRelease(t)
    gX|[gJB
    :Parameters
      :OpenSOAPTransportPtr [out] ((|t|))
        OpenSOAP Transport pointer
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportRelease(OpenSOAPTransportPtr /* [in, out] */ t) {
    int ret = OpenSOAPObjectRelease((OpenSOAPObjectPtr)t);

    return ret;
}

/*
=begin
--- function#OpenSOAPTransportSendHeader(t, content_length)
    Transport Header M

    :Parameters
      :OpenSOAPTransportPtr  [in] ((|t|))
        OpenSOAP Transport pointer
      :size_t [in] ((|content_length|))
        content length. if 0, then not add Content-Length Header
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
static
int
OpenSOAPTransportSendHeader(OpenSOAPTransportPtr /* [in] */ t,
                            size_t /* [in] */ content_length) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t && t->transportStream) {
        ret = OpenSOAPTransportHeaderSetContentLength(t->requestHeaders,
                                                      content_length);
        if (OPENSOAP_SUCCEEDED(ret)) {
            OpenSOAPByteArrayPtr http_hdr = NULL;
            ret = OpenSOAPByteArrayCreate(&http_hdr);

            if (OPENSOAP_SUCCEEDED(ret)) {
                ret = OpenSOAPConnectInfoGetHTTPPOSTRequest(t->serviceConnectInfo,
                                                            NULL,
                                                            http_hdr);
                if (OPENSOAP_SUCCEEDED(ret)) {
                    ret = OpenSOAPTransportHeaderAppendToHTTPHeader(t->requestHeaders,
                                                                    http_hdr);
                    
                    if (OPENSOAP_SUCCEEDED(ret)) {
                        ret = OpenSOAPStreamWriteByteArray(t->transportStream,
                                                           http_hdr,
                                                           NULL);
                    }
                }

                OpenSOAPByteArrayRelease(http_hdr);
            }
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportSend(t, soap_env)
    ByteArray ƂSOAPbZ[WMB

    :Parameters
      :OpenSOAPTransportPtr  [in] ((|t|))
        OpenSOAP Transport pointer
      :OpenSOAPByteArrayPtr [in] ((|soap_env|))
        SOAP Envelope
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
/* extern */
static
int
/* OPENSOAP_API */
OpenSOAPTransportSendByteArray(OpenSOAPTransportPtr /* [in] */ t,
                               OpenSOAPByteArrayPtr /* [in] */ soap_env) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t && t->transportStream && soap_env) {
        const unsigned char *soap_env_beg = NULL;
        size_t soap_env_sz = 0;
        /* get address and size */
        ret = OpenSOAPByteArrayGetBeginSizeConst(soap_env,
                                                 &soap_env_beg,
                                                 &soap_env_sz);
        if (OPENSOAP_SUCCEEDED(ret)) {
            /* Transport Header M */
            ret = OpenSOAPTransportSendHeader(t, soap_env_sz);
            if (OPENSOAP_SUCCEEDED(ret)) {
                /* Envelope Data M */
                ret = OpenSOAPStreamWrite(t->transportStream,
                                          soap_env_beg,
                                          &soap_env_sz);
            }
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportSend(t, soap_env)
    SOAPbZ[WM

    :Parameters
      :OpenSOAPTransportPtr  [in] ((|t|))
        OpenSOAP Transport pointer
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        SOAP Envelope
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportSend(OpenSOAPTransportPtr /* [in] */ t,
                      OpenSOAPEnvelopePtr /* [in] */ soap_env) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t && t->transportStream && soap_env) {
        OpenSOAPByteArrayPtr soap_env_data = NULL;
        ret = OpenSOAPByteArrayCreate(&soap_env_data);
        if (OPENSOAP_SUCCEEDED(ret)) {
            const unsigned char *charset_beg = NULL;
            ret = OpenSOAPTransportHeaderGetCharsetBeginConst(t->requestHeaders,
                                                              &charset_beg);
            if (OPENSOAP_SUCCEEDED(ret)) {
                if (!charset_beg || !*charset_beg) {
                    charset_beg = "UTF-8";
                }
                /* Envelope -> ByteArray */
                ret = OpenSOAPEnvelopeGetCharEncodingString(soap_env,
                                                            charset_beg,
                                                            soap_env_data);
                if (OPENSOAP_SUCCEEDED(ret)) {
                    ret = OpenSOAPTransportSendByteArray(t, soap_env_data);
                }
            }
            OpenSOAPByteArrayRelease(soap_env_data);
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportReceiveHeader(t, content_length)
    Transport Header MB

    :Parameters
      :OpenSOAPTransportPtr  [in] ((|t|))
        OpenSOAP Transport pointer
      :size_t * [out] ((|content_length|))
        Content-Length value stored buffer.
        if NULL, then not return Content-Length value.
      :int * [out] ((|tp_status|))
        Transport status value stored buffer.
        if NULL, then not return HTTP status value.
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
static
int
OpenSOAPTransportReceiveHeader(OpenSOAPTransportPtr /* [in] */ t,
                               size_t * /* [out] */ content_length,
                               int * /* [out] */ tp_status) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t && t->transportStream) {
        OpenSOAPByteArrayPtr http_hdr = NULL;
        ret = OpenSOAPByteArrayCreate(&http_hdr);
        if (OPENSOAP_SUCCEEDED(ret)) {
            size_t dummy_content_length = 0;
            if (!content_length) {
                content_length = &dummy_content_length;
            }
            ret = OpenSOAPStreamReadToDelm(t->transportStream,
                                           HTTP_HEADER_TERM,
                                           HTTP_HEADER_TERM_SZ,
                                           http_hdr);
            if (OPENSOAP_SUCCEEDED(ret)) {
                ret = OpenSOAPTransportHeaderSetFromHTTPHeader(t->responseHeaders,
                                                               http_hdr,
                                                               tp_status);
                if (OPENSOAP_SUCCEEDED(ret)) {
                    ret = OpenSOAPTransportHeaderGetContentLength(t->responseHeaders,
                                                                  content_length);
                }
            }
            OpenSOAPByteArrayRelease(http_hdr);
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportReceiveByteArray(t, soap_env)
    SOAPbZ[WM

    :Parameters
      :OpenSOAPTransportPtr [in] ((|t|))
        OpenSOAP Transport pointer
      :OpenSOAPByteArrayPtr [out] ((|soap_env|))
        SOAP Envelope data stored buffer.
      :int * [out] ((|tp_status|))
        Transport status value stored buffer.
        if NULL, then not return HTTP status value.
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
/* extern */
static
int
/* OPENSOAP_API */
OpenSOAPTransportReceiveByteArray(OpenSOAPTransportPtr /* [in] */ t,
                                  OpenSOAPByteArrayPtr
                                  /* [out] */ soap_env,
                                  int * /* [out] */ tp_status) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t) {
        size_t content_length = 0;
        /* Transport Header M */
        ret = OpenSOAPTransportReceiveHeader(t,
                                             &content_length,
                                             tp_status);
        if (OPENSOAP_SUCCEEDED(ret)) {
            /* Transport body M */
            ret = OpenSOAPStreamReadWithSize(t->transportStream,
                                             content_length,
                                             soap_env);
        }
    }
    
    return ret;
}

static
int
OpenSOAPTransportCreateHTTPEnvelope(OpenSOAPTransportPtr
                                    /* [in] */ t,
                                    OpenSOAPByteArrayPtr
                                    /* [in] */ response,
                                    int /* [in] */ http_status,
                                    OpenSOAPEnvelopePtr *
                                    /* [out] */ soap_env) {
    static const wchar_t FAULT_FORMAT[]
        =
        L"<SOAP-ENV:Envelope "
        L"xmlns:SOAP-ENV=\"http://schemas.xmlsoap.org/soap/envelope/\" "
        L"SOAP-ENV:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">"
        L"<SOAP-ENV:Body>"
        L"<SOAP-ENV:Fault>"
        L"<faultcode>OPENSOAP:HTTP_ERROR</faultcode>"
        L"<faultstring>HTTP RESPONSE STATUS %d</faultstring>"
        L"<detail>"
		L"<![CDATA["
        L"%ls"
		L"]]>"
        L"</detail>"
        L"</SOAP-ENV:Fault>"
        L"</SOAP-ENV:Body>"
        L"</SOAP-ENV:Envelope>";

    int ret = OPENSOAP_PARAMETER_BADVALUE;
    if (t && response && soap_env) {
        const unsigned char *charset_beg = NULL;
        ret = OpenSOAPTransportHeaderGetCharsetBeginConst(t->responseHeaders,
                                                          &charset_beg);
        if (OPENSOAP_SUCCEEDED(ret)) {
			int is_xml = 0;
            if (!charset_beg || !*charset_beg) {
                charset_beg = "UTF-8";
            }
			ret = OpenSOAPTransportHeaderIsContentTypeXML(t->responseHeaders,
														  &is_xml);
			
            if (OPENSOAP_SUCCEEDED(ret) && is_xml) {
                /* normal */
                ret = OpenSOAPEnvelopeCreateCharEncoding(charset_beg,
                                                         response,
                                                         soap_env);
            }
            else {
                /* http error */
                OpenSOAPStringPtr detail_str = NULL;
                ret = OpenSOAPStringCreateWithCharEncodingString(charset_beg,
                                                                 response,
                                                                 &detail_str);
                if (OPENSOAP_SUCCEEDED(ret)) {
                    size_t detail_sz = 0;
                    ret = OpenSOAPStringGetLengthWC(detail_str,
                                                    &detail_sz);
                    if (OPENSOAP_SUCCEEDED(ret)) {
                        wchar_t *detail = malloc(sizeof(wchar_t) *
                                                 (++detail_sz));
                        if (detail) {
                            ret = OpenSOAPStringGetStringWC(detail_str,
                                                            &detail_sz,
                                                            detail);
                            if (OPENSOAP_SUCCEEDED(ret)) {
                                OpenSOAPStringPtr res_str = NULL;
                                ret = OpenSOAPStringCreate(&res_str);
                                if (OPENSOAP_SUCCEEDED(ret)) {
                                    ret = OpenSOAPStringFormatWC(res_str,
                                                                 FAULT_FORMAT,
                                                                 http_status,
                                                                 detail);
                                    if (OPENSOAP_SUCCEEDED(ret)) {
                                        OpenSOAPByteArrayPtr res_buf = NULL;
                                        ret = OpenSOAPByteArrayCreate(&res_buf);
                                        if (OPENSOAP_SUCCEEDED(ret)) {
                                            ret = OpenSOAPStringGetCharEncodingString(res_str,
                                                                                      charset_beg,
                                                                                      res_buf);
                                            
                                            if (OPENSOAP_SUCCEEDED(ret)) {
                                                ret = OpenSOAPEnvelopeCreateCharEncoding(charset_beg,
                                                                                         res_buf,
                                                                                         soap_env);
                                            }
                                            OpenSOAPByteArrayRelease(res_buf);
                                        }
                                    }
                                    OpenSOAPStringRelease(res_str);
                                }
                            }

                            free(detail);
                        }
                        else {
                            ret = OPENSOAP_MEM_BADALLOC;
                        }
                    }
                    OpenSOAPStringRelease(detail_str);
                }
            }
        }
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPTransportReceive(t, soap_env)
    SOAPbZ[WM

    :Parameters
      :OpenSOAPTransportPtr [in] ((|t|))
        OpenSOAP Transport pointer
      :OpenSOAPEnvelopePtr * [in, out] ((|soap_env|))
        SOAP Envelope
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportReceive(OpenSOAPTransportPtr /* [in] */ t,
                         OpenSOAPEnvelopePtr * /* [in, out] */ soap_env) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t) {
        OpenSOAPByteArrayPtr soap_env_data = NULL;
        ret = OpenSOAPByteArrayCreate(&soap_env_data);
        if (OPENSOAP_SUCCEEDED(ret)) {
            int tp_status = 0;
            /* Envelope ByteArray M */
            ret = OpenSOAPTransportReceiveByteArray(t,
                                                    soap_env_data,
                                                    &tp_status);
            if (OPENSOAP_SUCCEEDED(ret)) {
                /* ByteArray -> Envelope */
                ret = OpenSOAPTransportCreateHTTPEnvelope(t,
                                                          soap_env_data,
                                                          tp_status,
                                                          soap_env);
            }

            OpenSOAPByteArrayRelease(soap_env_data);
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportConnect(t)
    End Point ւ̐ڑB

    :Parameters
      :OpenSOAPTransportPtr [in] ((|t|))
        OpenSOAP Transport pointer
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportConnect(OpenSOAPTransportPtr /* [in] */ t) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t) {
        OpenSOAPStreamPtr s = NULL;
        ret = OpenSOAPConnectInfoCreateClientSocketStream(t
                                                          ->serviceConnectInfo,
                                                          &s);
        if (OPENSOAP_SUCCEEDED(ret)) {
            OpenSOAPStreamRelease(t->transportStream);
            t->transportStream = s;
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportDisconnect(t)
    End Point ̐ؒfB

    :Parameters
      :OpenSOAPTransportPtr [in] ((|t|))
        OpenSOAP Transport pointer
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportDisconnect(OpenSOAPTransportPtr /* [in] */ t) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t) {
        OpenSOAPStreamRelease(t->transportStream);
        t->transportStream = NULL;
        ret = OPENSOAP_NO_ERROR;
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportInvokeByteArray(t, request, response)
    OpenSOAPByteArray ɂ SOAP ďoB

    :Parameters
      :OpenSOAPTransportPtr [in] ((|t|))
        OpenSOAP Transport pointer
      :OpenSOAPEnvelopePtr [in] ((|request|))
        SOAP request data.
      :OpenSOAPByteArrayPtr [out] ((|response|))
        SOAP response data.
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportInvokeByteArray(OpenSOAPTransportPtr /* [in] */ t,
                                 OpenSOAPByteArrayPtr /* [in] */ request,
                                 OpenSOAPByteArrayPtr /* [out] */ response,
                                 int * /* [out] */ tp_status) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t && request && response) {
        ret = OpenSOAPTransportConnect(t);
        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPTransportSendByteArray(t, request);
            if (OPENSOAP_SUCCEEDED(ret)) {
                ret = OpenSOAPTransportReceiveByteArray(t, response,
                                                        tp_status);
            }

            OpenSOAPTransportDisconnect(t);
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportInvoke(t, request, response)
    SOAPďoB

    :Parameters
      :OpenSOAPTransportPtr [in] ((|t|))
        OpenSOAP Transport pointer
      :OpenSOAPEnvelopePtr [in] ((|request|))
        SOAP request Envelope
      :OpenSOAPEnvelopePtr * [in, out] ((|response|))
        SOAP response Envelope
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportInvoke(OpenSOAPTransportPtr /* [in] */ t,
                        OpenSOAPEnvelopePtr /* [in] */ request,
                        OpenSOAPEnvelopePtr * /* [in, out] */ response) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t && request && response) {
        ret = OpenSOAPTransportConnect(t);
        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPTransportSend(t, request);
            if (OPENSOAP_SUCCEEDED(ret)) {
                ret = OpenSOAPTransportReceive(t, response);
            }

            OpenSOAPTransportDisconnect(t);
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportSetSOAPAction(s, soap_action)
    SOAP-Actionݒ(Some-URI)
    
    :Parameters
      :OpenSOAPTransportPtr [in, out] ((|t|))
        OpenSOAP Transport pointer
      :const char * [in] ((|soap_action|))
        soap-action.
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportSetSOAPAction(OpenSOAPTransportPtr /* [in, out] */ t,
                               const char * /* [in] */ soap_action) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (t) {
        if (!soap_action) {
            soap_action = "";
        }
        ret = OpenSOAPTransportHeaderSetHeader(t->requestHeaders,
                                               SOAPACTION_NAME,
                                               SOAPACTION_NAME_SZ,
                                               soap_action,
                                               strlen(soap_action));
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportSetURL(t, url)
    NGXgURLݒ(http://www.opensoap.com/StockQuote)

    :Parameters
      :OpenSOAPTransportPtr * [in, out] ((|t|))
        OpenSOAP Transport pointer
      :const char * [in] ((|url|))
        URL
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportSetURL(OpenSOAPTransportPtr /* [in, out] */ t,
                        const char * /* [in] */ url) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;
#if 1
    if (t && url && *url) {
        OpenSOAPConnectInfoPtr c_info = NULL;
        ret = OpenSOAPConnectInfoCreate(&c_info);
        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPConnectInfoSetURL(c_info, url);
            if (OPENSOAP_SUCCEEDED(ret)) {
                /* check valid protocol */
                int is_valid = 0;
                ret = OpenSOAPConnectInfoIsSameProtocol(c_info,
                                                        "http", &is_valid);
                if (OPENSOAP_SUCCEEDED(ret)) {
                    if (!is_valid) {
                        ret = OPENSOAP_NOT_CATEGORIZE_ERROR;
                    }
                    if (OPENSOAP_SUCCEEDED(ret)) {
                        ret = OpenSOAPByteArraySetData(t->serviceEndPoint,
                                                       url,
                                                       strlen(url) + 1);
                        if (OPENSOAP_SUCCEEDED(ret)) {
                            /* swap */
                            OpenSOAPConnectInfoPtr tmp = c_info;
                            c_info = t->serviceConnectInfo;
                            t->serviceConnectInfo = tmp;
                        }
                    }
                }
            }
            
            OpenSOAPConnectInfoRelease(c_info);
        }
    }
#else    

    if (t && url && *url) {
	OpenSOAPStringPtr soapurl;
	int beg_idx;
	int end_idx;
	int len;

	/* url𕪉 */
	ret = OpenSOAPStringCreateWithMB(url, &soapurl);
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}
	/* :܂łoB */
	ret = OpenSOAPStringFindStringMB(soapurl, ":", &beg_idx);
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}
	
	ret = OpenSOAPStringGetSubstring(soapurl, 0, beg_idx, &t->service);
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}

	/* //܂łoB */
	ret = OpenSOAPStringFindStringMB(soapurl, "//", &beg_idx);
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}
	/* /܂łoB */
	beg_idx += 2;
	end_idx = beg_idx;
	ret = OpenSOAPStringFindStringMB(soapurl, "/", &end_idx);
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}
	ret = OpenSOAPStringGetSubstring(soapurl, beg_idx, end_idx - beg_idx, &t->servername);
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}
	beg_idx = end_idx + 1;
	/* /܂łoB */
	ret = OpenSOAPStringGetLengthMB(soapurl, &len);
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}
	len -= beg_idx;
	ret = OpenSOAPStringGetSubstring(soapurl, beg_idx, len, &t->filename);
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}
    }

#endif    
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportSetHeader(t, header_name, header_value)
    wb_[ݒ

    :Parameters
      :OpenSOAPTransportPtr * [out] ((|t|))
        OpenSOAP Transport pointer
      :const char * [in] ((|header_name|))
        header name.
      :const char * [in] ((|header_value|))
        header's value.
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportSetHeader(OpenSOAPTransportPtr /* [in, out] */ t,
                           const char * /* [in] */ header_name,
                           const char * /* [in] */ header_value) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    if (t && header_name && *header_name) {
        size_t val_sz = (header_value) ? strlen(header_value) : 0;
	/* StringHashɕǉ */
        ret = OpenSOAPTransportHeaderSetHeader(t->requestHeaders,
                                               header_name,
                                               strlen(header_name),
                                               header_value,
                                               val_sz);
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPTransportSetCharset(s)
    LN^Zbgݒ

    :Parameters
      :OpenSOAPTransportPtr [in, out] ((|t|))
        OpenSOAP Transport pointer
      :const char * [in] charset
        character set.
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportSetCharset(OpenSOAPTransportPtr /* [out] */ t,
                            const char * /* [in] */ charset) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

#if 0
    if (t && charset && *charset) {
        ret = OpenSOAPTransportHeaderSetCharset(t->requestHeaders,
                                                charset);
    }
#else        
    static const wchar_t CONTENT_TYPE_FORMAT[]
        = L"text/xml; charset=\""
        STRING_WFORMAT
        L"\"";
    
    if (t && charset && *charset) {
        OpenSOAPStringPtr tmp = NULL;
        ret = OpenSOAPStringCreate(&tmp);
        if (OPENSOAP_SUCCEEDED(ret)) {
            OpenSOAPByteArrayPtr content_type = NULL;
            ret = OpenSOAPByteArrayCreate(&content_type);
            if (OPENSOAP_SUCCEEDED(ret)) {
                ret = OpenSOAPStringFormatWC(tmp,
                                             CONTENT_TYPE_FORMAT,
                                             charset);
                if (OPENSOAP_SUCCEEDED(ret)) {
                    ret = OpenSOAPStringGetCharEncodingString(tmp,
                                                              "US-ASCII",
                                                              content_type);
                    if (OPENSOAP_SUCCEEDED(ret)) {
                        const unsigned char *content_type_beg = NULL;
                        size_t content_type_sz = 0;
                        ret = OpenSOAPByteArrayGetBeginSizeConst(content_type,
                                                                 &content_type_beg,
                                                                 &content_type_sz);
                        if (OPENSOAP_SUCCEEDED(ret)) {
                            ret = OpenSOAPTransportHeaderSetHeader(t->requestHeaders,
                                                                   CONTENT_TYPE_NAME,
                                                                   sizeof(CONTENT_TYPE_NAME) - 1,
                                                                   content_type_beg,
                                                                   content_type_sz);
                        }
                    }
                }
                OpenSOAPByteArrayRelease(content_type);
            }

            OpenSOAPStringRelease(tmp);
        }
    }
#endif
    
    return ret;
}

/*
=begin
--- function#OpenSOAPTransportSMTPSetHostname(s)
    SMTPzXgݒ

    :Parameters
      :OpenSOAPTransportPtr * [out] ((|s|))
        OpenSOAP Transport pointer
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportSMTPSetHostname(OpenSOAPTransportPtr /* [out] */ s) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    return ret;
}

/*
=begin
--- function#OpenSOAPTransportSMTPSetFrom(t)
    SMTPMҐݒ

    :Parameters
      :OpenSOAPTransportPtr [out] ((|t|))
        OpenSOAP Transport pointer
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportSMTPSetFrom(OpenSOAPTransportPtr /* [out] */ t) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    return ret;
}

/*
=begin
--- function#OpenSOAPTransportSMTPSetTo(t)
    SMTPĐݒ

    :Parameters
      :OpenSOAPTransportPtr * [out] ((|t|))
        OpenSOAP Transport pointer
    :Returns
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPTransportSMTPSetTo(OpenSOAPTransportPtr /* [out] */ t) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    return ret;
}
